// @flow
import React, { Component } from 'react';
import {View, Image, TouchableOpacity, StatusBar, TouchableWithoutFeedback, Platform, Text} from 'react-native';
import {observable, action, computed, autorun, makeObservable, runInAction} from 'mobx';
import { observer } from 'mobx-react';
import { ColorSchemeContext, contextualStyleSheet, contextualValue, isDarkMode } from 'react-native-contextual-dark-mode';
import EntypoIcon from 'react-native-vector-icons/Entypo';
import { SafeAreaView, initialWindowMetrics } from 'react-native-safe-area-context';
import Video from 'react-native-video';
import compareVersions from 'compare-versions';

import { PinchZoomCarousel } from './PinchZoomCarousel';
import { ImageWithSpinner } from './ImageWithSpinner';
import { VideoController, VideoInformation } from './VideoController';
import { calculateSeekerPosition } from './VideoPlayerHelper';

import ArrowLeft from '../assets/arrowLeft.svg';
import ArrowRight from '../assets/arrowRight.svg';
import TrashIcon from '../assets/Trash.svg';
import EditIcon from '../assets/edit.svg';
import DownloadIcon from '../assets/download.svg';

const AndroidOSVersion = Platform.constants.Release;
const isVideoPreviewAvailable = !(
    Platform.OS === 'android' && compareVersions.compare(AndroidOSVersion, '7.0', '<')
);

export type PreviewMedia = {
    id: string;
    type: 'video' | 'image' | 'render';
    uri: string;
    onDeletePress?: () => void;
    onSaveMediaPress?: () => void;
    onEditPress?: () => void;
    onReplacePress?: () => void;
    extra: any;
}

export type MediaPreviewCarouselProps = {
    medias: PreviewMedia[];
    initialIndex?: number;
    renderMediaExtra?: (media: PreviewMedia) => any;
    onDragDown?: () => void;
    /** if you set a `imageSizeExtractor`, it avoids calling Image.getSize every time the picture is focused */
    imageSizeExtractor?: (media: PreviewMedia) => ?{ width: number, height: number };
    errorTitle: string;
    errorMessage: string;
    statusBarStyle?: 'light-content' | 'dark-content';
    /** override the style of the carousel */
    style?: any;
    onIndexChanged?: (newIndex: number) => void;
    unavailableVideoPreviewMessage?: string;
}

type JointMediaAndVideoInformation = {
    media: PreviewMedia,
    videoInformation?: VideoInformation, // Invariant: media.type === 'video' implies videoInformation is set
}

@observer
class MediaPreviewCarousel extends Component<MediaPreviewCarouselProps> {
    static contextType = ColorSchemeContext;
    static defaultProps = { initialIndex: 0 };

    // view state
    @observable showControls = true;
    @observable mediaHeight = null;
    @observable mediaWidth = null;
    @observable index = this.props.initialIndex;
    editorDisplayed = false;
    lastImageTapTime = 0;
    pinchZoomRef: ?PinchZoomCarousel = null;
    @observable seekerWidth = 0;
    @observable videoInformation: { [id: string]: VideoInformation } = {};
    videoRefs: { [id: string]: Video } = {};
    @observable scrollEnabled = true;

    constructor(props) {
        super(props);
        makeObservable(this);
        this.validateMedias();
        this.props.medias.forEach((previewMedia: PreviewMedia) => {
            if (previewMedia.type === 'video') {
                this.videoInformation[previewMedia.id] = {
                    duration: 0,
                    currentTime: 0,
                    paused: true,
                    seekerFillWidth: 0,
                    seekerPosition: 0,
                    seeking: false,
                    endReached: false,
                    seekerOffset: 0,
                    target: undefined,
                };
            }
        });
        autorun(this.readMediaDimensions);
    }

    // eslint-disable-next-line camelcase
    UNSAFE_componentWillReceiveProps(nextProps: MediaPreviewCarouselProps): * {
        if (this.props.medias.length > nextProps.medias.length) {
            // When deleting one or multiple media, move the index back by the amount deleted.
            // e.g. if you have 3 pictures and delete the middle one (index=1), the new index will be 0
            this.setIndex(Math.max(0, this.index - (this.props.medias.length - nextProps.medias.length)));
        }
    }

    validateMedias() {
        if (this.props.medias.length === 0) {
            throw new Error('There must be at least one media to show');
        }
        const mediaIds = new Set();
        this.props.medias.forEach((previewMedia: PreviewMedia) => {
            if (!previewMedia.id) {
                throw new Error('Each item in `medias` prop must have an id');
            }
            if (mediaIds.has(previewMedia.id)) {
                throw new Error('Duplicate id in `medias` prop is not supported');
            }
            mediaIds.add(previewMedia.id);
        });
    }

    @action
    setIndex = (newIndex: number) => {
        this.index = newIndex;
    }

    @computed
    get jointMediaAndVideoInformation(): JointMediaAndVideoInformation[] {
        return this.props.medias.map((media: PreviewMedia) => ({
            media,
            videoInformation: this.videoInformation[media.id],
        }));
    }

    @action
    updateVideoInformation = (mediaId: string, payload: VideoInformation) => {
        this.videoInformation[mediaId] = { ...this.videoInformation[mediaId], ...payload };
    };

    onProgressClosure = (mediaId: string) => action((data = {}) => {
        const { seeking, duration } = this.videoInformation[mediaId];
        if (!seeking) {
            this.videoInformation[mediaId].currentTime = data.currentTime;
            const position = calculateSeekerPosition(data.currentTime, duration, this.seekerWidth);
            this.setSeekerPosition(position, mediaId);
        }
    });

    onLoadClosure = (mediaId: string) => action((data: { duration: number, target: ?number }) => {
        if (this.videoInformation[mediaId].currentTime > 0) {
            this.videoRefs[mediaId].seek(this.videoInformation[mediaId].currentTime);
        }
        this.videoInformation[mediaId].duration = data.duration;
        this.videoInformation[mediaId].target = data.target;
    });

    onEndClosure = (mediaId: string) => action(() => {
        this.videoInformation[mediaId] = {
            ...this.videoInformation[mediaId],
            currentTime: 0,
            paused: true,
            seekerFillWidth: 0,
            seekerPosition: 0,
            seeking: false,
            endReached: true,
        };
        if (Platform.OS === 'android') {
            this.videoRefs[mediaId].seek(0);
        }
    });

    onPlaybackRateChangeClosure = (mediaId: string) => action(({ playbackRate, target }) => {
        const videoInfo = this.videoInformation[mediaId];
        // detect <Video> changing state playing -> paused on its own
        if (playbackRate === 0 && target === videoInfo.target && !videoInfo.seeking && !videoInfo.paused) {
            this.videoInformation[mediaId] = { ...videoInfo, paused: true };
        }
    });

    @action
    setSeekerPosition = (position: number, mediaId: string) => {
        const pos = Math.max(0, Math.min(this.seekerWidth, position));
        this.videoInformation[mediaId].seekerFillWidth = pos;
        this.videoInformation[mediaId].seekerPosition = pos;
        if (!this.videoInformation[mediaId].seeking) {
            this.videoInformation[mediaId].seekerOffset = position;
        }
    };

    @action
    setSeekerWidth = (width) => {
        if (width !== this.seekerWidth) {
            this.seekerWidth = width;
        }
    }

    @computed get mediaShown(): PreviewMedia {
        return this.props.medias[Math.max(0, Math.min(this.index, this.props.medias.length - 1))];
    }

    @computed
    get isPreviousButtonVisible(): boolean {
        return this.showControls && this.index > 0;
    }

    @computed
    get isNextButtonVisible(): boolean {
        return this.showControls && this.index < this.props.medias.length - 1;
    }

    attachPinchZoomRef = (ref) => {
        this.pinchZoomRef = ref;
    };

    @action
    pauseAllVideos = () => {
        Object.keys(this.videoInformation).forEach((key) => {
            this.videoInformation[key].paused = true;
        });
    };

    @action
    handleFocusIndexChanged = (newIndex: number) => {
        this.pauseAllVideos();
        if (newIndex >= this.props.medias.length || newIndex < 0) {
            // eslint-disable-next-line no-console
            console.warn(`Received new index ${newIndex} while there are ${this.props.medias.length} medias`, 'handleFocusIndexChanged');
        }
        const oldIndex = this.index;
        this.index = Math.max(0, Math.min(newIndex, this.props.medias.length - 1));
        this.props.onIndexChanged?.(this.index);
    };

    @action
    handleVideoTap = () => {
        this.showControls = !this.showControls;
    }

    @action
    handleImageTap = () => {
        const now = Date.now();
        if ((now - this.lastImageTapTime) < 300) {
            // double tap
            this.lastImageTapTime = 0;
            this.pinchZoomRef?.doubleTap();
        } else {
            this.lastImageTapTime = now;
            this.showControls = !this.showControls;
        }
    };

    @action
    setShowControls = (showControls) => {
        this.showControls = showControls;
    }

    @action
    updateHeightWidthMedia = (height, width) => {
        this.mediaHeight = height;
        this.mediaWidth = width;
    }

    goToNext = () => this.pinchZoomRef?.incrementIndex();

    goToPrevious = () => this.pinchZoomRef?.decrementIndex();

    @action
    hideButtons = () => {
        this.showControls = false;
    };

    @action
    showButtons = () => {
        this.showControls = true;
    };

    readMediaDimensions = () => {
        if (this.mediaShown.type === 'image') {
            if (this.props.imageSizeExtractor) {
                const result = this.props.imageSizeExtractor(this.mediaShown);
                if (result) {
                    this.updateHeightWidthMedia(result.height, result.width);
                    return;
                }
            }
            Image.getSize(
                this.mediaShown.uri,
                (width, height) => this.updateHeightWidthMedia(height, width),
                (error) => {
                    console.log('readMediaDimensions error:', error);
                    this.updateHeightWidthMedia(null, null);
                },
            );
        } else {
            this.updateHeightWidthMedia(null, null);
        }
    }

    hasNotchedTopScreen(): boolean {
        return initialWindowMetrics.insets.top > 20;
    }

    shouldHideStatusBar = (): boolean => {
        return !(Platform.OS === 'ios' && this.hasNotchedTopScreen()) || !this.showControls;
    }

    carouselKeyExtractor = ({ media }: JointMediaAndVideoInformation) => media.id;

    getScrollEnabled = () => this.scrollEnabled;
    getShowControls = () => this.showControls;

    /** Render methods */

    renderEditIcon = () => (
        <View style={s.optionBtnContainer}>
            <TouchableOpacity onPress={this.mediaShown.onEditPress} testID="MediaPreviewScreenEditBtn">
                <EditIcon width={FOOTER_ICON} height={FOOTER_ICON} fill={colors.footerIconOutline} />
            </TouchableOpacity>
        </View>
    );

    renderTrashIcon = () => (
        <View style={s.optionBtnContainer}>
            <TouchableOpacity testID="MediaPreviewScreenDelete" onPress={this.mediaShown.onDeletePress}>
                <TrashIcon width={FOOTER_ICON} height={FOOTER_ICON} fill={colors.footerIconOutline} />
            </TouchableOpacity>
        </View>
    );

    renderDownloadIcon = () => (
        <View style={s.optionBtnContainer}>
            <TouchableOpacity testID="MediaPreviewScreenDownload" onPress={this.mediaShown.onSaveMediaPress}>
                <DownloadIcon width={FOOTER_ICON} height={FOOTER_ICON} fill={colors.footerIconOutline} />
            </TouchableOpacity>
        </View>
    );

    renderReplaceBtn = () => (
        <View style={s.optionBtnContainer}>
            <TouchableOpacity onPress={this.mediaShown.onReplacePress} testID="MediaPreviewScreenReplaceBtn">
                <EntypoIcon size={FOOTER_ICON} name="retweet" color={colors.footerIconOutline} />
            </TouchableOpacity>
        </View>
    );

    renderPictureFooter = () => {
        if (!this.mediaShown.onReplacePress && !this.mediaShown.onEditPress && !this.mediaShown.onDeletePress) {
            return null;
        }
        return (
            <SafeAreaView style={s.footerContainer} edges={['bottom']} mode="padding">
                {!!this.mediaShown.onReplacePress && this.renderReplaceBtn()}
                {!!this.mediaShown.onEditPress && this.renderEditIcon()}
                {!!this.mediaShown.onSaveMediaPress && this.renderDownloadIcon()}
                {!!this.mediaShown.onDeletePress && this.renderTrashIcon()}
            </SafeAreaView>
        );
    }

    renderVideoFooterControls = () => (
        <VideoController
            id={this.mediaShown.id}
            updateVideoInformation={this.updateVideoInformation}
            videoInformation={this.videoInformation[this.mediaShown.id]}
            onSeek={(time: number) => { this.videoRefs[this.mediaShown.id].seek(time); }}
            isDarkMode={isDarkMode()}
            seekerWidth={this.seekerWidth}
            setSeekerWidth={this.setSeekerWidth}
            actionLeftCallback={this.mediaShown.onSaveMediaPress}
            actionLeftIcon={this.mediaShown.onSaveMediaPress ? (
                <DownloadIcon width={FOOTER_ICON} height={FOOTER_ICON} fill={isDarkMode() ? 'white' : 'black'} />
            ) : null}
            actionRightCallback={this.mediaShown.onDeletePress}
            actionRightIcon={this.mediaShown.onDeletePress ? (
                <TrashIcon width={FOOTER_ICON} height={FOOTER_ICON} fill={isDarkMode() ? 'white' : 'black'} />
            ) : null}
            onEnd={this.onEndClosure(this.mediaShown.id)}
        />
    );

    renderFooter = () => {
        if (['image', 'render'].includes(this.mediaShown.type)) {
            return this.renderPictureFooter();
        }
        if (this.mediaShown.type === 'video' && isVideoPreviewAvailable) {
            return this.renderVideoFooterControls();
        }
        return null;
    };

    renderPreviousBtn = () => (
        <SafeAreaView style={s.previousBtnContainer} edges={['left']}>
            <TouchableOpacity onPress={this.goToPrevious} style={s.previousBtn} testID="MediaPreviewScreenPreviousBtn">
                <ArrowLeft height={SMALL_ICON} width={SMALL_ICON} fill={colors.footerIconOutline} />
            </TouchableOpacity>
        </SafeAreaView>
    );

    renderNextBtn = () => (
        <SafeAreaView style={s.nextBtnContainer} edges={['right']}>
            <TouchableOpacity onPress={this.goToNext} style={s.nextBtn} testID="MediaPreviewScreenNextBtn">
                <ArrowRight height={SMALL_ICON} width={SMALL_ICON} fill={colors.footerIconOutline} />
            </TouchableOpacity>
        </SafeAreaView>
    );

    renderUnavailableVideoPreview = () => (
        <View style={s.errorOpenVideoContainer}>
            <Text style={s.errorOpenVideoText}>
                {this.props.unavailableVideoPreviewMessage || 'Video preview is not available'}
            </Text>
        </View>
    )

    renderMediaSubview = ({ media, videoInformation }: JointMediaAndVideoInformation, index: number, isFocused: boolean, _disableScroll: () => void, _enableScroll: () => void) => {
        const updateScroll = () => setImmediate(this.scrollEnabled ? _enableScroll : _disableScroll);
        updateScroll();
        const setScrollEnable = (scrollEnabled) => runInAction(() => {
            if (!isFocused) { return; }
            this.scrollEnabled = scrollEnabled;
            updateScroll();
        });

        if (media.type === 'image') {
            return (
                <ImageWithSpinner
                    onPress={this.handleImageTap}
                    imageUri={media.uri}
                    errorTitle={this.props.errorTitle}
                    errorMessage={this.props.errorMessage}
                />
            );
        }
        if (media.type === 'video') {
            return (
                <TouchableWithoutFeedback onPress={this.handleVideoTap}>
                    <View style={s.flex1}>
                        {isVideoPreviewAvailable ?
                            (this.index === index || Platform.OS === 'ios') && (
                                <Video
                                    ref={(video) => {
                                        this.videoRefs[media.id] = video;
                                    }}
                                    resizeMode="contain"
                                    // videoInformation.paused is not tracked by mobx, because
                                    // renderMediaSubview is called via a FlatList, outside the mobx tracking context
                                    paused={videoInformation.paused}
                                    muted={false}
                                    rate={1}
                                    onProgress={this.onProgressClosure(media.id)}
                                    onLoad={this.onLoadClosure(media.id)}
                                    onEnd={this.onEndClosure(media.id)}
                                    style={s.flex1}
                                    source={{ uri: media.uri }}
                                    ignoreSilentSwitch="ignore"
                                    onPlaybackRateChange={this.onPlaybackRateChangeClosure(media.id)}
                                    progressUpdateInterval={500}
                                />
                            ) : this.renderUnavailableVideoPreview()
                        }
                    </View>
                </TouchableWithoutFeedback>
            );
        }
        if (media.type === 'render') {
            return media.render(isFocused, setScrollEnable, this.setShowControls, this.getScrollEnabled, this.getShowControls);
        }
        // eslint-disable-next-line no-console
        console.warn(`Unknown preview media type "${media.type}"`);
        return null;
    }

    renderMediaCarousel = (canUseBlackBackground: boolean) => (
        <PinchZoomCarousel
            key="MediaCarousel"
            ref={this.attachPinchZoomRef}
            zoomDisabled={this.mediaShown.type !== 'image'}
            pictureHeight={this.mediaHeight}
            pictureWidth={this.mediaWidth}
            onZoomStart={this.hideButtons}
            onZoomEnd={this.showButtons}
            style={this.props.style || (!this.showControls && canUseBlackBackground ? s.carouselBlackBackground : s.carouselSchemeBackground)}
            onDragDown={this.props.onDragDown}
            items={this.jointMediaAndVideoInformation}
            index={this.index}
            renderItem={this.renderMediaSubview}
            keyExtractor={this.carouselKeyExtractor}
            onIndexChange={this.handleFocusIndexChanged}
        />
    );

    render() {
        const canUseBlackBackground = ['image', 'video'].includes(this.jointMediaAndVideoInformation?.[this.index]?.media?.type);
        return (
            <View style={s.screenContainer}>
                <StatusBar
                    hidden={this.shouldHideStatusBar()}
                    barStyle={this.props.statusBarStyle || (isDarkMode() ? 'light-content' : 'dark-content')}
                    animated
                />
                {this.renderMediaCarousel(canUseBlackBackground)}
                {this.showControls && this.renderFooter()}
                {this.showControls && this.props.renderMediaExtra?.(this.mediaShown)}
                {this.isPreviousButtonVisible && this.renderPreviousBtn()}
                {this.isNextButtonVisible && this.renderNextBtn()}
            </View>
        );
    }
}

const FOOTER_HEIGHT = 80;
const SMALL_ICON = 18;
const FOOTER_ICON = 30;

const colors = contextualValue((darkMode) => ({
    footerIconOutline: darkMode ? '#FFFFFF' : '#000000',
}));

const s = contextualStyleSheet((darkMode) => {
    const buttonBackground = darkMode ? '#1D252DB3' : '#EDEDEDB3';
    return {
        screenContainer: {
            flex: 1,
            flexDirection: 'row',
        },
        flex1: {
            flex: 1,
        },
        carouselBlackBackground: {
            flex: 1,
            backgroundColor: 'black',
        },
        carouselLightBackground: {
            flex: 1,
            backgroundColor: 'white',
        },
        carouselSchemeBackground: {
            flex: 1,
            backgroundColor: darkMode ? 'black' : 'white',
        },
        nextBtnContainer: {
            position: 'absolute',
            right: 15,
            alignSelf: 'center',
        },
        nextBtn: {
            padding: 9,
            backgroundColor: buttonBackground,
            borderRadius: 7,
        },
        previousBtnContainer: {
            position: 'absolute',
            left: 15,
            alignSelf: 'center',
        },
        previousBtn: {
            padding: 9,
            backgroundColor: buttonBackground,
            borderRadius: 7,
        },
        footerContainer: {
            position: 'absolute',
            bottom: 0,
            width: '100%',
            flexDirection: 'row',
            justifyContent: 'center',
            backgroundColor: buttonBackground,
        },
        optionBtnContainer: {
            flex: 1,
            alignItems: 'center',
            justifyContent: 'center',
            height: FOOTER_HEIGHT,
            minWidth: FOOTER_HEIGHT,
        },
        errorOpenVideoContainer: {
            flex: 1,
            alignItems: 'center',
            justifyContent: 'center',
            padding: 50,
        },
        errorOpenVideoText: {
            color: isDarkMode ? '#A1A3A6' : '#5B6770',
            fontSize: 17,
            textAlign: 'center',
        },
    };
});

export { MediaPreviewCarousel };
